@@ -1,11 +1,13 @@
 """
-Copyright CNRS/Inria/UniCA
+Copyright CNRS (https://www.cnrs.fr/index.php/en)
 Contributor(s): Eric Debreuve (eric.debreuve@cnrs.fr) since 2023
 SEE COPYRIGHT NOTICE BELOW
 """
 
 import dataclasses as d
+import inspect as e
 import logging as l
+import multiprocessing as prll
 import sys as s
 import textwrap as text
 import threading as thrd
@@ -14,41 +16,48 @@ import types as t
 import typing as h
 from datetime import date as date_t
 from datetime import datetime as date_time_t
-from os import sep as FOLDER_SEPARATOR
+from logging.handlers import QueueHandler as queue_handler_t
+from logging.handlers import QueueListener as log_server_t
 from pathlib import Path as path_t
 from traceback import TracebackException as traceback_t
 
-from logger_36.catalog.config.optional import MISSING_RICH_MESSAGE, RICH_IS_AVAILABLE
+from logger_36.catalog.config.optional import (
+    MEMORY_MEASURE_ERROR,
+    MEMORY_MEASURE_IS_AVAILABLE,
+    MISSING_RICH_MESSAGE,
+    RICH_IS_AVAILABLE,
+)
 from logger_36.catalog.handler.console import console_handler_t
 from logger_36.catalog.handler.file import file_handler_t
+from logger_36.catalog.handler.memory import memory_handler_t, records_h
 from logger_36.config.issue import ISSUE_CONTEXT_END, ISSUE_CONTEXT_SEPARATOR
 from logger_36.config.message import (
     DATE_FORMAT,
-    ELAPSED_TIME_SEPARATOR,
     LONG_ENOUGH,
     TIME_FORMAT,
+    WHERE_SEPARATOR,
 )
-from logger_36.constant.generic import NOT_PASSED
-from logger_36.constant.handler import ANONYMOUS, HANDLER_KINDS, handler_codes_h
+from logger_36.constant.chronos import DATE_ORIGIN, DATE_TIME_ORIGIN
 from logger_36.constant.issue import ISSUE_LEVEL_SEPARATOR, ORDER, order_h
-from logger_36.constant.logger import (
-    LOGGER_NAME,
-    WARNING_LOGGER_NAME,
-    WARNING_TYPE_COMPILED_PATTERN,
-)
+from logger_36.constant.logger import WARNING_LOGGER_NAME, WARNING_TYPE_COMPILED_PATTERN
 from logger_36.constant.memory import UNKNOWN_MEMORY_USAGE
-from logger_36.constant.message import LINE_INDENT, TIME_LENGTH_m_1, expected_op_h
-from logger_36.constant.path import PROJECT_FILE_RELATIVE
+from logger_36.constant.message import LINE_INDENT, expected_op_h
+from logger_36.constant.path import USER_FOLDER, LAUNCH_ROOT_FILE_relative
 from logger_36.constant.record import (
-    HIDE_WHERE_ATTR,
+    HAS_ACTUAL_EXPECTED_ATTR,
     SHOW_W_RULE_ATTR,
-    STORE_MEMORY_ATTR,
+    SHOW_WHEN_ATTR,
+    SHOW_WHERE_ATTR,
+    WHEN_OR_ELAPSED_ATTR,
+    WHERE_ATTR,
 )
-from logger_36.task.format.message import MessageFromRecord, MessageWithActualExpected
-from logger_36.task.format.rule import RuleAsText
-from logger_36.task.measure.chronos import ElapsedTime
+from logger_36.extension.file import NewTemporaryFile
+from logger_36.extension.sentinel import NOT_PASSED
+from logger_36.task.format.message import MessageWithActualExpected
+from logger_36.task.measure.chronos import FormattedElapsedTime
 from logger_36.task.measure.memory import CurrentUsage as CurrentMemoryUsage
-from logger_36.type.handler import any_handler_t as base_handler_t
+from logger_36.type.handler import extension_t as handler_extension_t
+from logger_36.type.handler import handler_h as base_handler_h
 from logger_36.type.issue import NewIssue, issue_t
 
 if RICH_IS_AVAILABLE:
@@ -64,8 +73,7 @@ logger_handle_raw_h = h.Callable[[l.LogRecord], None]
 logger_handle_with_self_h = h.Callable[[l.Logger, l.LogRecord], None]
 logger_handle_h = logger_handle_raw_h | logger_handle_with_self_h
 
-_DATE_TIME_ORIGIN = date_time_t.fromtimestamp(1970, None)
-_DATE_ORIGIN = _DATE_TIME_ORIGIN.date()
+MAIN_PROCESS_NAME = "MainProcess"
 
 
 @d.dataclass(slots=True, repr=False, eq=False)
@@ -74,44 +82,64 @@ class logger_t(base_t):
     intercepted_wrn_handle: When warning interception is on, this stores the original
         "handle" method of the Python warning logger.
 
-    _should_activate_log_interceptions: Loggers instantiated after a logger_t logger will
-    be missed by an early call of ToggleLogInterceptions. Therefore, passing True for
-    activate_log_interceptions only sets _should_activate_log_interceptions to True,
-    which is later checked in AddHandler to effectively call ToggleLogInterceptions if
-    _should_hold_messages is False (which normally indicates that the handler about to be
-    added is the last one).
-
-    _should_hold_messages: Must not be False until at least one handler has been added.
+    _should_activate_log_interceptions: Loggers instantiated after a logger_t logger
+    will be missed by an early call of ToggleLogInterceptions. Therefore, passing True
+    for activate_log_interceptions only sets _should_activate_log_interceptions to True,
+    which is later checked in AddHandler to effectively call ToggleLogInterceptions.
     """
 
-    should_record_messages: bool = False
     exit_on_error: bool = False  # Implies exit_on_critical.
     exit_on_critical: bool = False
+    should_monitor_memory_usage: bool = False
 
-    events: dict[int, int] = d.field(init=False, default_factory=dict)
-    recorded: list[tuple[int, str]] = d.field(init=False, default_factory=list)
+    history: dict[date_time_t, str] = d.field(init=False, default_factory=dict)
+    n_events: dict[int, int] = d.field(init=False, default_factory=dict)
 
-    last_message_now: date_time_t = d.field(init=False, default=_DATE_TIME_ORIGIN)
-    last_message_date: date_t = d.field(init=False, default=_DATE_ORIGIN)
-    any_handler_stores_memory: bool = d.field(init=False, default=False)
+    last_message_now: date_time_t = d.field(init=False, default=DATE_TIME_ORIGIN)
+    last_message_date: date_t = d.field(init=False, default=DATE_ORIGIN)
     memory_usages: list[tuple[str, int]] = d.field(init=False, default_factory=list)
     context_levels: list[str] = d.field(init=False, default_factory=list)
-    staged_issues: list[issue_t] = d.field(init=False, default_factory=list)
+    staged_issues: list[tuple[issue_t, bool]] = d.field(
+        init=False, default_factory=list
+    )
     intercepted_wrn_handle: logger_handle_h | None = d.field(init=False, default=None)
     intercepted_log_handles: dict[str, logger_handle_h] = d.field(
         init=False, default_factory=dict
     )
     intercepts_exceptions: bool = d.field(init=False, default=False)
-    _on_hold: list[l.LogRecord] = d.field(init=False, default_factory=list)
-    _should_hold_messages: bool = d.field(init=False, default=True)
+
+    # Used only until the first handler is added (see AddHandler).
     _should_activate_log_interceptions: bool = d.field(init=False, default=False)
 
-    name_: d.InitVar[str] = LOGGER_NAME
+    log_server: log_server_t | None = d.field(init=False, default=None)
+
+    name_: d.InitVar[str | None] = None
     level_: d.InitVar[int] = l.NOTSET
     activate_wrn_interceptions: d.InitVar[bool] = True
     activate_log_interceptions: d.InitVar[bool] = True
     activate_exc_interceptions: d.InitVar[bool] = True
 
+    @property
+    def formatted_history(self) -> str:
+        """"""
+        FormattedEntry = lambda _: f"{_[0]}: {_[1].replace('\n', '↲ ')}"
+        return "\n".join(map(FormattedEntry, self.history.items()))
+
+    @property
+    def records(self) -> records_h | None:
+        """"""
+        return logger_t.Records(self)
+
+    @staticmethod
+    def Records(logger: base_t | l.Logger, /) -> records_h | None:
+        """"""
+        for handler in logger.handlers:
+            output = getattr(handler, "records", None)
+            if memory_handler_t.AreRecords(output):
+                return output
+
+        return None
+
     @property
     def intercepts_warnings(self) -> bool:
         """"""
@@ -152,21 +180,31 @@ class logger_t(base_t):
 
     def __post_init__(
         self,
-        name_: str,
+        name_: str | None,
         level_: int,
         activate_wrn_interceptions: bool,
         activate_log_interceptions: bool,
         activate_exc_interceptions: bool,
     ) -> None:
         """"""
+        assert prll.current_process().name == MAIN_PROCESS_NAME
+
+        if name_ is None:
+            name_ = f"{type(self).__name__}:{hex(id(self))[2:]}"
+
+        self.history[date_time_t.now()] = (
+            f'Logger "{name_}" instantiation for "{LAUNCH_ROOT_FILE_relative}"'
+        )
+
         base_t.__init__(self, name_)
         self.setLevel(level_)
         self.propagate = False  # Part of base_t.
 
-        for level in l.getLevelNamesMapping().values():
-            self.events[level] = 0
+        if self.exit_on_error:
+            self.exit_on_critical = True
 
-        self.info(f'New logger "{self.name}" for "{PROJECT_FILE_RELATIVE}"')
+        for level_id in l.getLevelNamesMapping().values():
+            self.n_events[level_id] = 0
 
         if activate_wrn_interceptions:
             self.ToggleWarningInterceptions(True)
@@ -175,153 +213,67 @@ class logger_t(base_t):
         if activate_exc_interceptions:
             self.ToggleExceptionInterceptions(True)
 
-        if self.exit_on_error:
-            self.exit_on_critical = True
+        if self.should_monitor_memory_usage:
+            self.ActivateMemoryUsageMonitoring()
 
     def handle(self, record: l.LogRecord, /) -> None:
         """"""
-        elapsed_time, now = ElapsedTime(should_return_now=True)
-
-        if (self._on_hold.__len__() > 0) and not self._should_hold_messages:
-            for held in self._on_hold:
-                if self.should_record_messages:
-                    self.recorded.append(
-                        (held.levelno, MessageFromRecord(held, RuleAsText)[0])
-                    )
-                base_t.handle(self, held)
-            self._on_hold.clear()
-
+        now = date_time_t.now()
         if (date := now.date()) != self.last_message_date:
-            self.last_message_date = date
-            # levelno: Added for management by logging.Logger.handle.
-            date_record = l.makeLogRecord(
-                {
-                    "name": self.name,
-                    "levelno": l.INFO,
-                    "msg": f"DATE: {date.strftime(DATE_FORMAT)}",
-                    SHOW_W_RULE_ATTR: True,
-                }
-            )
-            if self._should_hold_messages:
-                self._on_hold.append(date_record)
-            else:
-                if self.should_record_messages:
-                    self.recorded.append(
-                        (
-                            date_record.levelno,
-                            MessageFromRecord(date_record, RuleAsText)[0],
-                        )
-                    )
-                base_t.handle(self, date_record)
+            self._AcknowledgeDateChange(date)
+
+        level = record.levelno
 
         # When.
-        if now - self.last_message_now > LONG_ENOUGH:
-            record.when_or_elapsed = now.strftime(TIME_FORMAT)
-        else:
-            record.when_or_elapsed = (
-                f"{ELAPSED_TIME_SEPARATOR}{elapsed_time:.<{TIME_LENGTH_m_1}}"
-            )
+        if getattr(record, SHOW_WHEN_ATTR, True):
+            if now - self.last_message_now > LONG_ENOUGH:
+                w_or_e = f"{now:{TIME_FORMAT}}"
+            else:
+                w_or_e = FormattedElapsedTime(now)  # or: f"{[...]:.<{TIME_LENGTH}}".
+            setattr(record, WHEN_OR_ELAPSED_ATTR, w_or_e)
         self.last_message_now = now
 
         # Where.
-        # Memory usage is also stored if there are no handlers yet, just in case.
-        should_store_where = self.any_handler_stores_memory or not self.hasHandlers()
-        should_show_where = (record.levelno != l.INFO) and not hasattr(
-            record, HIDE_WHERE_ATTR
-        )
-        if should_store_where or should_show_where:
-            module = path_t(record.pathname)
-            for path in s.path:
-                if module.is_relative_to(path):
-                    module = module.relative_to(path).with_suffix("")
-                    module = str(module).replace(FOLDER_SEPARATOR, ".")
-                    break
-            else:
-                module = record.module
-            where = f"{module}:{record.funcName}:{record.lineno}"
+        should_show_where = getattr(record, SHOW_WHERE_ATTR, level != l.INFO)
+        if should_show_where or self.should_monitor_memory_usage:
+            where = f"{record.pathname}:{record.funcName}:{record.lineno}"
             if should_show_where:
-                record.where = where
-        else:
-            where = None
-
-        # How.
-        record.level_first_letter = record.levelname[0]
+                setattr(record, WHERE_ATTR, where)
+            if self.should_monitor_memory_usage:
+                self.memory_usages.append((where, CurrentMemoryUsage()))
 
         # What.
         if not isinstance(record.msg, str):
             record.msg = str(record.msg)
 
-        if self._should_hold_messages:
-            self._on_hold.append(record)
-        else:
-            if self.should_record_messages:
-                self.recorded.append(
-                    (record.levelno, MessageFromRecord(record, RuleAsText)[0])
-                )
-            base_t.handle(self, record)
+        base_t.handle(self, record)
+        self.n_events[level] += 1
 
-        if (self.exit_on_critical and (record.levelno is l.CRITICAL)) or (
-            self.exit_on_error and (record.levelno is l.ERROR)
+        if (self.exit_on_critical and (level is l.CRITICAL)) or (
+            self.exit_on_error and (level is l.ERROR)
         ):
             # Also works if self.exit_on_error and record.levelno is l.CRITICAL since
             # __post_init__ set self.exit_on_critical if self.exit_on_error.
             s.exit(1)
 
-        self.events[record.levelno] += 1
-
-        if should_store_where:
-            self.memory_usages.append((where, CurrentMemoryUsage()))
-
-    def SetLevel(self, level: int, /, *, which: handler_codes_h | str = "a") -> None:
-        """
-        Set level of handlers, but the logger level is not modified.
-
-        which: if not a handler_codes_h, then it corresponds to a handler name.
-        """
-        found = False
-        for handler in self.handlers:
-            if (
-                (which == "a")
-                or ((which in "cfg") and (getattr(handler, "kind", None) == which))
-                or (which == handler.name)
-            ):
-                handler.setLevel(level)
-                if which not in HANDLER_KINDS:
-                    return
-                found = True
-
-        if not found:
-            raise ValueError(
-                MessageWithActualExpected(
-                    "Handler not found",
-                    actual=which,
-                    expected=f"{str(HANDLER_KINDS)[1:-1]}, or a handler name",
-                )
-            )
-
-    def MakeMonochrome(self) -> None:
+    def _AcknowledgeDateChange(self, date: date_t, /) -> None:
         """"""
-        self.AddHandler(console_handler_t)
+        self.last_message_date = date
 
-    def MakeRich(self, *, alternating_logs: int = 0) -> None:
-        """"""
-        if MISSING_RICH_MESSAGE is not None:
-            s.__stderr__.write(MISSING_RICH_MESSAGE + "\n")
-
-        if console_rich_handler_t is console_handler_t:
-            handler_kwargs = {}
-        else:
-            handler_kwargs = {"alternating_logs": alternating_logs}
-        self.AddHandler(console_rich_handler_t, **handler_kwargs)
-
-    def MakePermanent(self, path: str | path_t, /) -> None:
-        """"""
-        self.AddHandler(file_handler_t, path=path)
+        record = l.makeLogRecord(
+            {
+                "name": self.name,
+                "levelno": l.INFO,  # For management by logging.Logger.handle.
+                "msg": f"DATE: {date:{DATE_FORMAT}}",
+                SHOW_W_RULE_ATTR: True,
+            }
+        )
+        base_t.handle(self, record)
 
     def ResetEventCounts(self) -> None:
         """"""
-        for level in self.events:
-            self.events[level] = 0
+        for level_id in self.n_events:
+            self.n_events[level_id] = 0
 
     def ToggleWarningInterceptions(self, state: bool, /) -> None:
         """"""
@@ -334,7 +286,7 @@ class logger_t(base_t):
             logger.handle = t.MethodType(_HandleForWarnings(self), logger)
 
             l.captureWarnings(True)
-            self.info("Warning Interception: ON")
+            self.history[date_time_t.now()] = "Warning Interception: ON"
         else:
             if not self.intercepts_warnings:
                 return
@@ -344,12 +296,12 @@ class logger_t(base_t):
             self.intercepted_wrn_handle = None
 
             l.captureWarnings(False)
-            self.info("Warning Interception: OFF")
+            self.history[date_time_t.now()] = "Warning Interception: OFF"
 
     def ToggleLogInterceptions(self, state: bool, /) -> None:
         """"""
         if state:
-            if self.intercepts_logs:
+            if self._should_activate_log_interceptions or self.intercepts_logs:
                 return
 
             # Note: Alternative to self.manager is logging.root.manager.
@@ -368,8 +320,14 @@ class logger_t(base_t):
             intercepted = sorted(self.intercepted_log_handles.keys())
             if intercepted.__len__() > 0:
                 as_str = ", ".join(intercepted)
-                self.info(f"Now Intercepting LOGs from: {as_str}")
+                self.history[date_time_t.now()] = (
+                    f"Now Intercepting LOGs from: {as_str}"
+                )
         else:
+            if self._should_activate_log_interceptions:
+                self._should_activate_log_interceptions = False
+                return
+
             if not self.intercepts_logs:
                 return
 
@@ -377,7 +335,7 @@ class logger_t(base_t):
                 logger = l.getLogger(name)
                 logger.handle = handle
             self.intercepted_log_handles.clear()
-            self.info("Log Interception: OFF")
+            self.history[date_time_t.now()] = "Log Interception: OFF"
 
     def ToggleExceptionInterceptions(self, state: bool, /) -> None:
         """"""
@@ -388,7 +346,7 @@ class logger_t(base_t):
             s.excepthook = self.DealWithException
             thrd.excepthook = self.DealWithExceptionInThread
             self.intercepts_exceptions = True
-            self.info("Exception Interception: ON")
+            self.history[date_time_t.now()] = "Exception Interception: ON"
         else:
             if not self.intercepts_exceptions:
                 return
@@ -396,74 +354,111 @@ class logger_t(base_t):
             s.excepthook = s.__excepthook__
             thrd.excepthook = thrd.__excepthook__
             self.intercepts_exceptions = False
-            self.info("Exception Interception: OFF")
+            self.history[date_time_t.now()] = "Exception Interception: OFF"
 
-    def DealWithException(self, _, exc_value, exc_traceback, /) -> None:
+    def ActivateMemoryUsageMonitoring(self) -> None:
         """"""
-        exception = exc_value.with_traceback(exc_traceback)
-        self.LogException(exception, level=l.CRITICAL)
-        s.exit(1)
-
-    def DealWithExceptionInThread(
-        self, exc_type, exc_value, exc_traceback, _, /
-    ) -> None:
-        """"""
-        self.DealWithException(exc_type, exc_value, exc_traceback)
+        if MEMORY_MEASURE_IS_AVAILABLE:
+            # Useless if called from __post_init__.
+            self.should_monitor_memory_usage = True
+            self.history[date_time_t.now()] = (
+                f'Memory usage monitoring activated for logger "{self.name}"'
+            )
+        else:
+            self.should_monitor_memory_usage = False
+            self.error(MEMORY_MEASURE_ERROR)
 
     def AddHandler(
         self,
-        handler_t: type[base_handler_t],
+        handler_t_or_handler: type[base_handler_h]
+        | base_handler_h
+        | l.Handler
+        | l.FileHandler,
         /,
         *,
         name: str | None = None,
         level: int = l.INFO,
-        should_store_memory_usage: bool = False,
         message_width: int = -1,
-        formatter: l.Formatter | None = None,
-        should_still_hold_messages: bool = False,
         **kwargs,
     ) -> None:
-        """"""
-        if (not should_still_hold_messages) and self._should_activate_log_interceptions:
-            self.ToggleLogInterceptions(True)
+        """
+        Silently ignores re-holding request after un-holding.
+        """
+        if self._should_activate_log_interceptions:
+            # Turn _should_activate_log_interceptions off before calling
+            # ToggleLogInterceptions because it checks it.
             self._should_activate_log_interceptions = False
+            self.ToggleLogInterceptions(True)
 
-        self._should_hold_messages = should_still_hold_messages
-        handler = handler_t.New(
-            name=name,
-            should_store_memory_usage=should_store_memory_usage,
-            message_width=message_width,
-            level=level,
-            formatter=formatter,
-            **kwargs,
-        )
-        base_t.addHandler(self, handler)
-
-        extension = getattr(handler, "extension", None)
-        if extension is None:
-            name = handler.name
-            if (name is None) or (name.__len__() == 0):
-                name = ANONYMOUS
+        if isinstance(handler_t_or_handler, type):
+            handler = handler_t_or_handler.New(
+                name=name, message_width=message_width, level=level, **kwargs
+            )
         else:
-            name = getattr(extension, "name", ANONYMOUS)
-            if getattr(extension, STORE_MEMORY_ATTR, False):
-                self.any_handler_stores_memory = True
+            handler = handler_t_or_handler
+        base_t.addHandler(self, handler)
 
         path = getattr(handler, "baseFilename", "")
         if isinstance(path, path_t) or (path.__len__() > 0):
             path = f"\nPath: {path}"
-
-        self.info(
-            f'New handler "{name}" of type "{type(handler).__name__}" and '
+        self.history[date_time_t.now()] = (
+            f'New handler "{handler.name}" of type "{type(handler).__name__}" and '
             f"level {handler.level}={l.getLevelName(handler.level)}{path}"
         )
 
+    def MakeMonochrome(self) -> None:
+        """"""
+        self.AddHandler(console_handler_t)
+
+    def MakeRich(self, *, alternating_logs: int = 0) -> None:
+        """"""
+        if RICH_IS_AVAILABLE:
+            handler_kwargs = {"alternating_logs": alternating_logs}
+        else:
+            handler_kwargs = {}
+            self.error(MISSING_RICH_MESSAGE)
+
+        self.AddHandler(console_rich_handler_t, **handler_kwargs)
+
+    def MakePermanent(self, path: str | path_t, /) -> None:
+        """"""
+        self.AddHandler(file_handler_t, path=path)
+
+    def MakeMultiSafe(self) -> None:
+        """
+        Should not be called until after all desired handlers have been added (as a
+        better-then-nothing check, it is checked that the logger has at least one
+        handler). If handlers are added passed this call, execution might freeze or
+        crash.
+        """
+        assert self.log_server is None
+        assert self.hasHandlers()
+
+        handlers = tuple(self.handlers)  # Making a copy is necessary.
+        for handler in handlers:
+            self.removeHandler(handler)
+
+        queue = prll.Queue()
+
+        self.addHandler(queue_handler_t(queue))
+
+        self.log_server = log_server_t(queue, *handlers)
+        self.log_server.start()
+
     def __call__(self, *args, **kwargs) -> None:
         """
-        For a print-like calling.
+        For a print-like calling for print-based debugging.
         """
         separator = kwargs.get("separator", " ")
-        self.info(separator.join(map(str, args)))
+
+        frame = e.stack(context=0)[1][0]  # 1=caller.
+        details = e.getframeinfo(frame, context=0)
+        path = path_t(details.filename)
+        if path.is_relative_to(USER_FOLDER):
+            path = path.relative_to(USER_FOLDER)
+        where = f"{str(path.with_suffix(''))}:{details.function}:{details.lineno}"
+
+        self.info(separator.join(map(str, args)) + f"\n{WHERE_SEPARATOR} " + where)
 
     def Log(
         self,
@@ -480,7 +475,7 @@ class logger_t(base_t):
         """"""
         if isinstance(level, str):
             level = l.getLevelNamesMapping()[level.upper()]
-        message = MessageWithActualExpected(
+        message, has_actual_expected = MessageWithActualExpected(
             message,
             actual=actual,
             expected=expected,
@@ -488,7 +483,25 @@ class logger_t(base_t):
             expected_op=expected_op,
             with_final_dot=with_final_dot,
         )
-        self.log(level, message)
+        if has_actual_expected:
+            extra = {HAS_ACTUAL_EXPECTED_ATTR: True}
+        else:
+            extra = {}
+        self.log(level, message, extra=extra)
+
+    def LogAsIs(self, message: str, /, *, indented: bool = False) -> None:
+        """"""
+        if indented:
+            message = text.indent(message, LINE_INDENT)
+
+        for handler in self.handlers:
+            EmitMessage = getattr(
+                handler, handler_extension_t.EmitMessage.__name__, None
+            )
+            if EmitMessage is not None:
+                EmitMessage(message)
+
+    info_raw = LogAsIs  # To follow the convention of the logging methods info, error...
 
     def LogException(
         self,
@@ -510,24 +523,26 @@ class logger_t(base_t):
             #     - Why adding exception name here and not when removing caller?
             formatted = "".join(lines)
             message = f"Exception of type {type(exception).__name__}\n----\n{formatted}"
-        self.log(level, message, extra={HIDE_WHERE_ATTR: None})
+        self.log(level, message, extra={SHOW_WHERE_ATTR: False})
 
-    def LogAsIs(self, message: str, /, *, indented: bool = False) -> None:
+    def DealWithException(self, _, exc_value, exc_traceback, /) -> None:
         """"""
-        if indented:
-            message = text.indent(message, LINE_INDENT)
+        exception = exc_value.with_traceback(exc_traceback)
+        self.LogException(exception, level=l.CRITICAL)
+        s.exit(1)
 
-        for handler in self.handlers:
-            handler.LogAsIs(message)
-
-    info_raw = LogAsIs  # To follow the convention of the logging methods info, error...
+    def DealWithExceptionInThread(self, args, /) -> None:
+        """"""
+        self.DealWithException(args.exc_type, args.exc_value, args.exc_traceback)
 
     def DisplayRule(
         self, /, *, message: str | None = None, color: str = "white"
     ) -> None:
         """"""
         for handler in self.handlers:
-            handler.DisplayRule(text=message, color=color)
+            EmitRule = getattr(handler, handler_extension_t.EmitRule.__name__, None)
+            if EmitRule is not None:
+                EmitRule(text=message, color=color)
 
     def AddContextLevel(self, new_level: str, /) -> None:
         """"""
@@ -569,7 +584,9 @@ class logger_t(base_t):
         )
         self.staged_issues.append(issue)
 
-    def PopIssues(self, /, *, should_remove_context: bool = False) -> list[str]:
+    def PopIssues(
+        self, /, *, should_remove_context: bool = False
+    ) -> list[tuple[str, bool]]:
         """"""
         if not self.has_staged_issues:
             return []
@@ -581,10 +598,10 @@ class logger_t(base_t):
         else:
             separator = ISSUE_LEVEL_SEPARATOR
         separator_length = separator.__len__()
-        for issue in self.staged_issues:
+        for issue, has_actual_expected in self.staged_issues:
             start_idx = issue.find(separator)
             issue = issue[(start_idx + separator_length) :]
-            output.append(issue)
+            output.append((issue, has_actual_expected))
 
         self.staged_issues.clear()
 
@@ -606,13 +623,16 @@ class logger_t(base_t):
                     "Invalid commit order",
                     actual=order,
                     expected=f"One of {str(ORDER)[1:-1]}",
-                )
+                )[0]
             )
 
         if order == "when":
             issues = self.staged_issues
         else:  # order == "context"
-            issues = sorted(self.staged_issues, key=lambda _: _.context)
+            issues = sorted(
+                self.staged_issues,
+                key=lambda _: _[0].split(ISSUE_LEVEL_SEPARATOR, maxsplit=1)[1],
+            )
         """
         Format issues as an exception:
         try:
@@ -624,20 +644,43 @@ class logger_t(base_t):
             formatted = "\n".join(lines)
         """
 
-        hide_where = {HIDE_WHERE_ATTR: None}
+        extra = {SHOW_WHERE_ATTR: False}
         if unified:
-            level, _ = issues[0].split(ISSUE_LEVEL_SEPARATOR, maxsplit=1)
+            level, _ = issues[0][0].split(ISSUE_LEVEL_SEPARATOR, maxsplit=1)
             wo_level = []
-            for issue in issues:
+            any_has_actual_expected = False
+            for issue, has_actual_expected in issues:
                 _, issue = issue.split(ISSUE_LEVEL_SEPARATOR, maxsplit=1)
+                if has_actual_expected:
+                    any_has_actual_expected = True
                 wo_level.append(issue)
-            self.log(int(level), "\n".join(wo_level), stacklevel=2, extra=hide_where)
+            if any_has_actual_expected:
+                extra[HAS_ACTUAL_EXPECTED_ATTR] = True
+            self.log(int(level), "\n".join(wo_level), stacklevel=2, extra=extra)
         else:
-            for issue in issues:
+            for issue, has_actual_expected in issues:
                 level, issue = issue.split(ISSUE_LEVEL_SEPARATOR, maxsplit=1)
-                self.log(int(level), issue, stacklevel=2, extra=hide_where)
+                if has_actual_expected:
+                    extra[HAS_ACTUAL_EXPECTED_ATTR] = True
+                self.log(int(level), issue, stacklevel=2, extra=extra)
+                if has_actual_expected:
+                    del extra[HAS_ACTUAL_EXPECTED_ATTR]
         self.staged_issues.clear()
 
+    def StoragePath(self, suffix: str, /) -> path_t:
+        """
+        Use as staticmethod if needed.
+        """
+        for handler in self.handlers:
+            if (path := getattr(handler, "baseFilename", None)) is not None:
+                output = path_t(path).with_suffix(suffix)
+                if output.exists():
+                    output = NewTemporaryFile(suffix)
+
+                return output
+
+        return NewTemporaryFile(suffix)
+
     def __enter__(self) -> None:
         """"""
         pass
@@ -653,6 +696,13 @@ class logger_t(base_t):
         _ = self.context_levels.pop()
         return False
 
+    def __del__(self) -> None:
+        """"""
+        assert prll.current_process().name == MAIN_PROCESS_NAME
+
+        if self.log_server is not None:
+            self.log_server.stop()
+
 
 def _HandleForWarnings(interceptor: base_t, /) -> logger_handle_h:
     """"""
@@ -730,18 +780,4 @@ The fact that you are presently reading this means that you have had
 knowledge of the CeCILL license and that you accept its terms.
 
 SEE LICENCE NOTICE: file README-LICENCE-utf8.txt at project source root.
-
-This software is being developed by Eric Debreuve, a CNRS employee and
-member of team Morpheme.
-Team Morpheme is a joint team between Inria, CNRS, and UniCA.
-It is hosted by the Centre Inria d'Université Côte d'Azur, Laboratory
-I3S, and Laboratory iBV.
-
-CNRS: https://www.cnrs.fr/index.php/en
-Inria: https://www.inria.fr/en/
-UniCA: https://univ-cotedazur.eu/
-Centre Inria d'Université Côte d'Azur: https://www.inria.fr/en/centre/sophia/
-I3S: https://www.i3s.unice.fr/en/
-iBV: http://ibv.unice.fr/
-Team Morpheme: https://team.inria.fr/morpheme/
 """
